from __future__ import annotations
from typing import Any

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.container import BarContainer
from matplotlib.figure import Figure

from maidr.core.enum.plot_type import PlotType
from maidr.core.maidr import Maidr
from maidr.core.maidr_plot_data_factory import MaidrPlotDataFactory


class FigureManager:
    """
    A helper class responsible for creating and associating `Maidr` objects with
    `Figure` objects.

    Methods
    -------
    create_maidr(axs, plot_types)
        Creates a `Maidr` object with associated plot data based on the plot and its
        type.
    get_axes(artist)
        Retrieves the `Axes` object associated with a given matplotlib `Artist`.

    Warnings
    --------
    End users will typically not have to use this class directly.
    """

    figs = dict()

    @classmethod
    def create_maidr(
        cls, axs: list[Axes | None], plot_types: list[PlotType]
    ) -> Maidr | tuple[Maidr]:
        if not axs:
            raise ValueError("No plots found.")
        if None in axs:
            raise ValueError("NoneType found in plots.")
        if plot_types is None:
            raise ValueError("No plot types found.")
        if len(axs) != len(plot_types):
            raise ValueError(
                f"Lengths of plots {len(axs)} and their {len(plot_types)} types do "
                "not match."
            )

        ordered_maidrs = list()
        visited_maidrs = set()

        for ax, plot_type in zip(axs, plot_types):
            fig = ax.get_figure()  # type: ignore
            if fig is None:
                raise ValueError(f"No figure found for axis: {ax}.")

            # maintain the order of unique MAIDRs as extracted from the axes
            maidr = cls.__get_maidr(fig)
            if fig not in visited_maidrs:
                ordered_maidrs.append(maidr)
                visited_maidrs.add(fig)

            # convert the plot to MAIDR representation
            maidr_data = MaidrPlotDataFactory.create(ax, plot_type)  # type: ignore
            maidr.add_plot(maidr_data)

        if len(ordered_maidrs) == 1:
            return ordered_maidrs[0]
        else:
            return tuple(*ordered_maidrs)

    @classmethod
    def __get_maidr(cls, fig: Figure) -> Maidr:
        if fig not in cls.figs.keys():
            cls.figs[fig] = Maidr(fig)
        return cls.figs[fig]

    @staticmethod
    def get_axes(plot: Any) -> list[Axes | None]:
        """
        Extracts axes from various types of matplotlib artists or collections thereof.
        Always returns a list of Axes objects, which may include None if an axes object
        cannot be found for a given artist.
        """
        axs = FigureManager.__get_axes(plot)
        # Ensure output is a flat list of Axes objects
        return axs if isinstance(axs, list) else [axs]

    @staticmethod
    def __get_axes(
        artist: Artist | Axes | BarContainer | list | np.ndarray | None,
    ) -> Any:
        """
        Recursively extracts Axes objects from the input artist or collection of artists.
        """
        if artist is None:
            return None
        elif isinstance(artist, Axes):
            return artist
        elif isinstance(artist, BarContainer):
            # Get axes from the first occurrence of any child artist
            return next(
                (
                    child_artist.axes
                    for child_artist in artist.get_children()
                    if isinstance(child_artist.axes, Axes)
                ),
                None,
            )
        elif isinstance(artist, Artist):
            return artist.axes
        elif isinstance(artist, (list, np.ndarray)):
            # Flatten list to avoid nested lists
            flat_list = []
            for a in artist:
                extracted_axes = FigureManager.__get_axes(a)
                if isinstance(extracted_axes, list):
                    flat_list.extend(extracted_axes)
                else:
                    flat_list.append(extracted_axes)
            return flat_list



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from maidr.core.enum.maidr_key import MaidrKey
from maidr.core.enum.plot_type import PlotType
from maidr.core.maidr_plot_data import MaidrPlotData
from maidr.exception.extraction_error import ExtractionError


class LinePlotData(MaidrPlotData):
    last_extracted_id = -1

    def __init__(self, axes: Axes) -> None:
        super().__init__(axes, PlotType.LINE)

    def _extract_maidr_data(self) -> dict:
        plt_type = self.type.value
        ax = self.axes

        maidr = {
            MaidrKey.TYPE.value: plt_type,
            MaidrKey.TITLE.value: ax.get_title(),
            MaidrKey.AXES.value: {
                MaidrKey.X.value: {
                    MaidrKey.LABEL.value: ax.get_xlabel(),
                },
                MaidrKey.Y.value: {
                    MaidrKey.LABEL.value: ax.get_ylabel(),
                },
            },
            MaidrKey.DATA.value: self.__extract_data(),
        }

        return maidr

    def __extract_data(self) -> list[dict]:
        ax = self.axes
        plot = None
        data = None

        if isinstance(ax, Axes):
            self.last_extracted_id += 1
            plot = LinePlotData.__extract_line(ax, self.last_extracted_id)
        if isinstance(plot, Line2D):
            data = LinePlotData.__extract_line_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    @staticmethod
    def __extract_line_data(plot: Line2D) -> list[dict] | None:
        if plot is None:
            return None

        return [
            {
                MaidrKey.X.value: float(x),
                MaidrKey.Y.value: float(y),
            }
            for x, y in plot.get_xydata()
        ]

    @staticmethod
    def __extract_line(plot: Axes, line_id: int) -> Line2D | None:
        if plot is None or plot.get_lines() is None:
            return None

        return plot.get_lines()[line_id]



from __future__ import annotations

import numpy as np
from matplotlib.axes import Axes
from matplotlib.collections import QuadMesh
from matplotlib.container import BarContainer
from matplotlib.image import AxesImage
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon
from numpy import ndarray

from maidr.core.enum.plot_type import PlotType
from maidr.core.maidr import Maidr
from maidr.core.figure_manager import FigureManager


def bar(
    plot: Axes | BarContainer | list[Axes | BarContainer] | ndarray,
) -> Maidr | tuple[Maidr]:
    """
    Create a Maidr object for a bar plot.

    Parameters
    ----------
    plot : Axes | BarContainer
        The bar plot for which a Maidr object is to be created.

    Returns
    -------
    Maidr
        The created Maidr object representing the bar plot.

    Raises
    ------
    ValueError
        If the input `plot` is missing the `matplotlib.figure.Figure` and
        `matplotlib.figure.Axes`.
    TypeError
        If the input `plot` is not a valid bar plot.

    See Also
    --------
    Maidr : The core class encapsulating the plot with its MAIDR structure.

    Examples
    --------
        >>> import matplotlib.pyplot as plt
        >>> import maidr
        >>> bar_plot = plt.bar(5, [1, 2, 3, 4, 5])  # Generate a bar plot
        >>> bar_maidr = maidr.bar(bar_plot)  # Convert the plot to a Maidr object
        >>> bar_maidr.save("maidr_bar_plot.html")  # Save the plot to an HTML file
    """
    axs = FigureManager.get_axes(plot)
    plot_type = [PlotType.BAR for _ in axs]
    return FigureManager.create_maidr(axs, plot_type)


def count(
    plot: Axes | BarContainer | list[Axes | BarContainer] | ndarray,
) -> Maidr | tuple[Maidr]:
    """
    Create a Maidr object for a count plot.

    Parameters
    ----------
    plot : Axes | list[Axes] | BarContainer | list[BarContainer]
        The count plot for which a Maidr object is to be created.

    Returns
    -------
    Maidr
        The created Maidr object representing the count plot.

    Raises
    ------
    ValueError
        If the input `plot` is missing the `matplotlib.figure.Figure` and
        `matplotlib.figure.Axes`.
    TypeError
        If the input `plot` is not a valid count plot.

    See Also
    --------
    Maidr : The core class encapsulating the plot with its MAIDR structure.
    bar : Function to create a Maidr object for matplotlib bar plots, usable as an
    alternative to `count()`.

    Note
    ----
    Since a count plot is a specific case of a bar plot, this function internally uses
    the `bar()` function to process the plot. The `count()` function is provided as a
    convenience to align with the `seaborn.countplot()` method.

    Examples
    --------
        >>> import seaborn as sns
        >>> import maidr
        >>> data = sns.load_dataset("titanic")  # Load the dataset
        >>> count_plot = sns.countplot(x="class", data=data)  # Generate a count plot
        >>> count_maidr = maidr.count(count_plot)  # Convert the plot to a Maidr object
        >>> count_maidr.save("maidr_count_plot.html")  # Save the plot to an HTML file
    """
    return bar(plot)


def heat(
    plot: Axes | AxesImage | QuadMesh | list[Axes | AxesImage | QuadMesh] | ndarray,
) -> Maidr | tuple[Maidr]:
    axs = FigureManager.get_axes(plot)
    plot_type = [PlotType.HEAT for _ in axs]
    return FigureManager.create_maidr(axs, plot_type)


def hist(
    plot: BarContainer | Polygon | list[BarContainer | Polygon] | np.ndarray,
) -> Maidr | tuple[Maidr]:
    axs = FigureManager.get_axes(plot)
    plot_type = [PlotType.HIST for _ in axs]
    return FigureManager.create_maidr(axs, plot_type)


def line(plot: Line2D | list[Line2D]) -> Maidr | tuple[Maidr]:
    axs = FigureManager.get_axes(plot)
    plot_type = [PlotType.LINE for _ in axs]
    return FigureManager.create_maidr(axs, plot_type)


def close() -> None:
    pass



from __future__ import annotations

import numpy as np
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable

from maidr.core.enum.maidr_key import MaidrKey
from maidr.core.enum.plot_type import PlotType
from maidr.core.maidr_plot_data import MaidrPlotData
from maidr.exception.extraction_error import ExtractionError


class HeatPlotData(MaidrPlotData):
    """
    A class encapsulating all the MAIDR representation of the axes in a figure
    along with the plot.

    This class extends `MaidrPlotData` to specifically handle data extraction and
    representation for heatmaps. It encapsulates the details required to represent a
    heatmap as a part of an interactive MAIDR visualization, including plot type,
    titles, axes labels, and the plot data itself.

    Parameters
    ----------
    axes : Axes
        The matplotlib axes object on which the heatmap is drawn.

    Warnings
    --------
    End users will typically not have to use this class directly.

    See Also
    --------
    MaidrPlotData : The base class for MAIDR plot data objects.
    """

    def __init__(self, axes: Axes) -> None:
        """
        Initializes the HeatPlotData object with matplotlib axes with matplotlib axes
        containing the heatmap.

        Parameters
        ----------
        axes : Axes
            The axes object associated with the heatmap.
        """
        super().__init__(axes, PlotType.HEAT)

    def _extract_maidr_data(self) -> dict:
        """
        Extracts and structures heatmap data for MAIDR visualization.

        Returns
        -------
        dict
            A dictionary containing the extracted maidr heatmap data.
        """
        plt_type = self.type.value
        ax = self.axes

        maidr = {
            MaidrKey.TYPE.value: plt_type,
            MaidrKey.TITLE.value: ax.get_title(),
            MaidrKey.LABEL.value: {
                MaidrKey.FILL.value: None,
            },
            MaidrKey.AXES.value: {
                MaidrKey.X.value: {
                    MaidrKey.LABEL.value: ax.get_xlabel(),
                    MaidrKey.LEVEL.value: self.__extract_x_level(),
                },
                MaidrKey.Y.value: {
                    MaidrKey.LABEL.value: ax.get_ylabel(),
                    MaidrKey.LEVEL.value: self.__extract_y_level(),
                },
            },
            MaidrKey.DATA.value: self.__extract_data(),
        }

        return maidr

    def __extract_x_level(self) -> list:
        """
        Extracts x-axis level values based on tick labels.

        Returns
        -------
        list
            A list of strings representing the x-axis levels.
        """
        return [label.get_text() for label in self.axes.get_xticklabels()]

    def __extract_y_level(self) -> list:
        """
        Extracts y-axis level values based on tick labels.

        Returns
        -------
        list
            A list of strings representing the y-axis levels.
        """
        return [label.get_text() for label in self.axes.get_yticklabels()]

    def __extract_data(self) -> list[list]:
        """
        Extracts numerical data from the heatmap.

        Returns
        -------
        list[list]
            A 2D list of numerical data extracted from the heatmap.

        Raises
        ------
        ExtractionError
            If the plot object is incompatible for data extraction.
        """
        ax = self.axes
        plot = None
        data = None

        if isinstance(ax, Axes):
            plot = HeatPlotData.__extract_scalar_mappable(ax)
        if isinstance(plot, ScalarMappable):
            data = HeatPlotData.__extract_scalar_mappable_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    @staticmethod
    def __extract_scalar_mappable_data(plot: ScalarMappable) -> list[list] | None:
        """
        Extracts numerical data from the specified ScalarMappable object if possible.

        Parameters
        ----------
        plot : ScalarMappable
            The QuadMesh from which to extract the data.

        Returns
        -------
        list[list] | None
            A 2D list containing the numerical data extracted from the ScalarMappable,
            or None if the plot does not contain valid data values or is not a
            ScalarMappable.
        """
        if plot is None or plot.get_array() is None:
            return None

        data = list()
        array = plot.get_array()
        for row in array:  # type: ignore
            row_data = list()
            for item in row:
                if isinstance(item, np.integer):
                    row_data.append(int())
                elif isinstance(item, np.floating):
                    row_data.append(float(item))
                else:
                    row_data.append(item)
            data.append(row_data)

        return data

    @staticmethod
    def __extract_scalar_mappable(plot: Axes) -> ScalarMappable | None:
        """
        Extracts the ScalarMappable from the given Axes object if possible.

        Parameters
        ----------
        plot : Axes
            The Axes object to search for a ScalarMappable.

        Returns
        -------
        ScalarMappable | None
            The first ScalarMappable found within the given Axes object, or None if no
            ScalarMappable is present.
        """
        # Ideally, there should only be one AxesImage/QuadMesh for a heatmap
        for child in plot.get_children():
            if isinstance(child, ScalarMappable):
                return child



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core.enum.maidr_key import MaidrKey
from maidr.core.enum.plot_type import PlotType
from maidr.core.maidr_plot_data import MaidrPlotData
from maidr.exception.extraction_error import ExtractionError


class HistPlotData(MaidrPlotData):
    def __init__(self, axes: Axes) -> None:
        """
        Initializes the HistPlotData object with matplotlib axes with matplotlib axes
        containing the histogram.

        Parameters
        ----------
        axes : Axes
            The axes object associated with the histogram.
        """
        super().__init__(axes, PlotType.HIST)

    def _extract_maidr_data(self) -> dict:
        """
        Extracts and structures histogram data for MAIDR visualization.

        Returns
        -------
        dict
            A dictionary containing the extracted maidr heatmap data.
        """
        plt_type = self.type.value
        ax = self.axes

        maidr = {
            MaidrKey.TYPE.value: plt_type,
            MaidrKey.TITLE.value: ax.get_title(),
            MaidrKey.AXES.value: {
                MaidrKey.X.value: {
                    MaidrKey.LABEL.value: ax.get_xlabel(),
                },
                MaidrKey.Y.value: {
                    MaidrKey.LABEL.value: ax.get_ylabel(),
                },
            },
            MaidrKey.DATA.value: self.__extract_data(),
        }

        return maidr

    def __extract_data(self) -> list[dict]:
        """
        Extracts numerical data from the histogram.

        Returns
        -------
        list[dict]
            A 2D list of numerical data extracted from the heatmap.

        Raises
        ------
        ExtractionError
            If the plot object is incompatible for data extraction.
        """
        ax = self.axes
        plot = None
        data = None

        if isinstance(ax, Axes):
            plot = HistPlotData.__extract_bar_container(ax)
        if isinstance(plot, BarContainer):
            data = HistPlotData.__extract_bar_container_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    @staticmethod
    def __extract_bar_container_data(plot: BarContainer) -> list[dict] | None:
        """
        Extracts numerical data from the specified ScalarMappable object if possible.

        Parameters
        ----------
        plot : BarContainer
            The BarContainer from which to extract the data.

        Returns
        -------
        list[dict] | None
            A 2D list containing the numerical data extracted from the BarContainer,
            or None if the plot does not contain valid data values or is not a
            BarContainer.
        """
        if plot is None or plot.patches is None:
            return None

        data = list()
        for patch in plot.patches:
            y = float(patch.get_height())
            x = float(patch.get_x())
            width = float(patch.get_width())
            data.append(
                {
                    MaidrKey.Y.value: y,
                    MaidrKey.X.value: x + width / 2,
                    MaidrKey.X_MIN.value: x,
                    MaidrKey.X_MAX.value: x + width,
                    MaidrKey.Y_MIN.value: 0,
                    MaidrKey.Y_MAX.value: y,
                }
            )

        return data

    @staticmethod
    def __extract_bar_container(plot: Axes) -> BarContainer | None:
        """
        Extracts the BarContainer from the given Axes object if possible.

        Parameters
        ----------
        plot : Axes
            The Axes object to search for a BarContainer.

        Returns
        -------
        BarContainer | None
            The first BarContainer found within the given Axes object, or None if no
            BarContainer is present.
        """
        for container in plot.containers:
            if isinstance(container, BarContainer):
                return container
